#!/usr/bin/ruby
DIR=File.dirname((
  f = __FILE__
  while File.lstat(f).symlink?
    f = "#{File.dirname f}/#{File.readlink f}".gsub(/^.*\/\//, "/")
  end
  f
))
require "#{DIR}/../config"
require "#{DIR}/common"
require "socket"
require "fileutils"

# deamonize (assuming setsid did the grunt work)
$stdin.force_close
$stdout.force_close
$stderr.force_close
$stdout = $stderr = File.open(LOG_FILE, 'w')

# kill the old daemon if applicable
begin
  Process.kill("SIGHUP", File.open(DAEMON_PID_FILE, &:read).to_i)
rescue Errno::ENOENT, Errno::EPERM, Errno::ESRCH
  # the daemon is not even started, pff!
end

# clean up the old daemon's data and meta-data
FileUtils.rm_rf(DAEMON_PID_FILE)
FileUtils.rm_rf(DAEMON_SOCKET_FILE)
FileUtils.rm_rf(DATA_DIR)
FileUtils.mkdir(DATA_DIR)

# prepare death ritual
socket_daemon = nil
trap("SIGHUP") {
  socket_daemon.force_close if socket_daemon
  $stdout.force_close
}

# advertize our location
File.open(DAEMON_PID_FILE, 'w') {|f|
  f.puts $$
}

# files are deemed unused when the creator's friends (same program group)
# are all dead
def pgids
  r = []
  Dir.foreach("/proc") {|filename|
    begin
      r << Process.getpgid(filename.split("/").last.to_i)
    rescue Errno::ESRCH
      # process died while looping, not a live pgid anymore
    end
  }
  r.sort.uniq
end

# now listen...
timeout = nil
pgid2file_list = {}
socket_daemon = (
  umask = File.umask
  begin
    File.umask 0
    UNIXServer.new DAEMON_SOCKET_FILE
  ensure
    File.umask umask
  end
)
loop {
  r = IO.select([socket_daemon], [], [], timeout)
  begin
    if r
      socket = socket_daemon.accept
      begin
        pgid = socket.gets.to_i
        pgid2file_list[pgid] ||= []
        timeout = 1
        filename = "#{DATA_DIR}/#{pgid}.#{pgid2file_list[pgid].length}.tmp"
        File.open(filename, 'w') {|f| f.write(socket.read)}
        pgid2file_list[pgid] << filename
        socket.puts filename
      ensure
        socket.force_close
      end
    elsif timeout
      maintained = {}
      pgids.each {|pgid|
        maintained[pgid] = pgid2file_list.delete(pgid) if pgid2file_list[pgid]
      }
      
      if pgid2file_list.empty?
        # gradually fall into slumber
        timeout *= 2
      else
        # unlink the files which aren't needed anymore
        pgid2file_list.each {|pgid, files|
          files.each {|filename|
            FileUtils.rm_rf(filename)
          }
        }
      end
      
      pgid2file_list = maintained
    end
  ensure
    $stdout.force_flush
  end
}
