require 'yaml'

module RPC
  class Proxy
    attr_reader :socket
    def initialize socket
      @socket = socket
    end
    
    def method_missing(method_name, *args)
      msg = {
        :method_name => method_name,
        :args => args
      }.to_yaml
      @socket.write([msg.length, msg].pack("Na*"))
      @socket.flush
      msg = YAML::load(@socket.read(@socket.read(4).unpack("N").first))
      raise msg[:exception], msg[:message] if msg[:exception]
      msg[:return_value]
    end
  end
  
  class Served_Object
    def self.served_method method_name
      @@served_methods ||= []
      @@served_methods << method_name
    end
    
    def serve socket
      msg = YAML::load(socket.read(socket.read(4).unpack("N").first))
      begin
        unless @@served_methods.include?(msg[:method_name])
          raise NoMethodError.new, "undefined method `#{msg[:method_name]}'"
        end
        msg = {
          :return_value => self.send(msg[:method_name], *msg[:args])
        }.to_yaml
        socket.write([msg.length, msg].pack("Na*"))
        socket.flush
      rescue
        msg = {
          :exception => $!,
          :message => $!.message
        }.to_yaml
        socket.write([msg.length, msg].pack("Na*"))
        socket.flush
      end
    end
  end
end
